/*
 * Ship name
 * Based on source files with separate class lists
 * by Roger Burton West <rburtonw@nyx.net>
 * GNUWare
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <tkplayer.h>

void help()
{
  fprintf(stderr,"Usage: shipname {<-f hullID|-r>} {-t} {-I path} <dir>\n");
  fprintf(stderr,"See documentation for more info!\n");
}

void makesrc(FILE *handle,int hullid,SHIPDATA *shipdata,HULLSPEC *hullspec) {
  fprintf(handle,"%d:%s\n",hullid,(const char *)hullspec[hullid].sName);
  if (shipdata) {
    for (int s=1;s<=500;s++) {
      if (shipdata[s].wID>0) {
        if (shipdata[s].wShipType==hullid) {
          fprintf(handle,"%s\n",(const char *)shipdata[s].sName);
        }
      }
    }
  }
}

int main(int argc,char *argv[])
{
  int hullforce=0;
  bool force=false;
  bool gensrc=false;
  bool usepath=false;
  bool trimlong=false;
  int c;
  int race;
  HULLSPEC *hullspec;
  TRUEHULL *truehull;
  SHIPDATA *shipdata;
  XString shfile;
  XString path;
  XString dir;
  FILE *handle;

  srandom(time(NULL));

  fprintf(stderr,"SHIPNAME v0.81 by Roger Burton West <roger@firedrake.demon.co.uk>\n");

  opterr = 0;
  optind=1;
  while ((c=getopt(argc, argv, "rf:htI:")) != -1) {
    switch (c) {
      case 'r':
        gensrc=true;
        break;
      case 'f':
        force=true;
        hullforce=atoi(optarg);
        break;
      case 'I':
        usepath=true;
        path=optarg;
        break;
      case 't':
        trimlong=true;
        break;
      case 'h':
        help();
        return 1;
      case '?':
        fprintf(stderr,"Unknown option %c - use -h for help\n", c);
        return 1;
    }
  }
  if (optind < argc) {
    set_game_dir(argv[optind]);
    dir=argv[optind];
  } else {
    fprintf(stderr,"Insufficient parameters - use -h for help\n");
    return 1;
  }

  hullspec=read_all_hullspec();
  if (!hullspec) {
    fprintf(stderr,"Couldn't open hullspec.dat\n");
    return 1;
  }

  for(race=1;race<=11;race++) {
    if (set_player(race)) {
      truehull=read_all_truehull();
      if (!truehull) {
        fprintf(stderr,"Couldn't open truehull.dat\n");
        return 1;
      }
      shipdata=read_all_shipdata();
      if (!shipdata && !gensrc) {
        fprintf(stderr,"Couldn't open ship%d.dat\n",race);
        return 1;
      }
      if (gensrc) {
        handle=mkopen("names&p.nm",false,'o');
        if (handle) {
          bool donehull[MAXHULL+1];
          fprintf(handle,"'This is a Ship Namer source file.\n");
          fprintf(handle,"'Comment lines begin with quotes.\n");
          fprintf(handle,"'Blank lines will be ignored.\n");
          fprintf(handle,"'Hull section header lines have the format hullID:name.\n");
          fprintf(handle,"'Name prefix lines should have the format -prefix.\n");
          fprintf(handle,"'Include file lines should have the format @file,\n");
          fprintf(handle,"' which may be followed by a (comma-sep) section specification.\n");
          fprintf(handle,"'Any other line will be interpreted as a ship name.\n");
          fprintf(handle,"'Ship Namer is copyleft 1996-8 by Roger Burton West.\n");
          fprintf(handle,"0:Overflow list\n");
          fprintf(handle,"\n'Native hulls:\n");
          for (int h=1;h<=MAXHULL;h++) {
            donehull[h]=false;
          }
          for (int h=1;h<21;h++) {
            if (truehull[h].wHullnum!=0) {
              donehull[truehull[h].wHullnum]=true;
              makesrc(handle,truehull[h].wHullnum,shipdata,hullspec);
            }
          }
          fprintf(handle,"\n'Captured/traded hulls:\n");
          for (int h=1;h<=MAXHULL;h++) {
            if (!donehull[h]) {
              if (hullspec[h].sName != "") {
                makesrc(handle,h,shipdata,hullspec);
              }
            }
          }
          fclose(handle);
        }
      } else {
        handle=mkopen("names&p.nm",true,'r');
        XString notused="\r";
        if (handle) {
          bool rlist[MAXHULL+1],slist[501];
          for (int s=0;s<=MAXHULL;s++) {
            rlist[s]=false;
          }
          for (int s=1;s<=500;s++) {
            slist[s]=false;
          }
          //iterate over all ships, get the ones to be renamed
          for (int s=1;s<=500;s++) {
            bool ca=false;
            slist[s]=false;
            if (shipdata[s].wID>0) {
              if (force) {
                if (shipdata[s].wShipType==hullforce) ca=true;
                if (hullforce==0) ca=true;
              } else {
                XString temp,temp2;
                temp=shipdata[s].sName;
                if (temp.Left(5)=="SHIP ") ca=true;
                temp2=hullspec[shipdata[s].wShipType].sName.Lower().Left(20);
                temp2.Trim();
                if (temp.Lower() == temp2) {
                  ca=true;
                }
                if (temp=="Rename") ca=true;
              }
              if (ca) {
                rlist[shipdata[s].wShipType]=true;
                slist[s]=true;
              }
            }
            if (!ca) { // not getting renamed, add to non-use list
              notused |= shipdata[s].sName;
              notused |= "\r";
            }
          }
          //read name lists
          XString prefix[MAXHULL+1];
          XString namelist[MAXHULL+1][300];
          XString apref;
          XString section="";
          int namesize[MAXHULL+1];
          int active=0;
          bool include=false;
          bool sectnow=true;
          bool trap=false;
          FILE *h2=NULL;

          for (int s=0;s<=MAXHULL;s++) namesize[s]=0;
          do {
            int ch;
            char line[100];
            XString temp,temp2;

            ch=-1;
            do {
              ch++;
              fread(&line[ch],1,1,handle);
            } while (line[ch]>=' ');
            do {
              fread(&line[ch],1,1,handle);
            } while (line[ch]<' ' && !feof(handle));

            if (include && feof(handle)) {
              include=false;
              trap=true;
              fclose(handle);
              handle=h2;
            } else {
              if (!feof(handle)) fseek(handle,-1,SEEK_CUR);
            }

            line[ch]=0;

            if (line[0]!='\'' && line[0]!=0) { //comment or blank
              if (strstr(line,":") != NULL) {
                active=atoi(line);
                apref=prefix[0];
                if (active>MAXHULL) active=0;
              } else {
                if (line[0]=='@') { // start include file
                  if (!include) {
                    XString fntemp,fnt2;
                    int ll;
                    fntemp=line;
                    fntemp=fntemp.Mid(1);
                    ll=fntemp.Find(' ');
                    if (ll>=0) {
                      section=-fntemp.Mid(ll+1);
                      section="," | section | ",";
                      fntemp=fntemp.Left(ll);
                      sectnow=false;
                    }
                    h2=handle;
                    fnt2=dir | "/" | fntemp;
                    handle=fopen((const char *)fnt2,"rb");
                    if (!handle) {
                      fnt2=fntemp;
                      handle=fopen((const char *)fnt2,"rb");
                    }
                    if (!handle && usepath) {
                      fnt2=path | "/" | fntemp;
                      handle=fopen((const char *)fnt2,"rb");
                    }
                    if (handle) {
                      include=true;
                    } else {
                      handle=h2;
                    }
                  }
                } else {
                  if (line[0]=='[') {
                    if (section!="") {
                      XString temp;
                      temp=line;
                      temp=-temp.Mid(1,temp.Len()-2);
                      temp="," | temp | ",";
                      sectnow=(section.Find((const char *)temp)>=0);
                    }
                  } else {
                    if ((rlist[active] || active==0) && sectnow) {
                      if (line[0]=='-') { // ship prefix
                        prefix[active]=line+1;
                        apref=prefix[active];
                      } else { // ship name
                        temp="";
                        if (apref != "") {
                          temp |= apref;
                          temp |= " ";
                        }
                        temp |= line;
                        temp2 = "\r" | temp | "\r";
                        if (notused.Find((const char *)temp2)==-1) {
                          if (trimlong) {
                            temp=temp.Left(20);
                          } else {
                            if (temp.Len()>20) temp="";
                          }
                          if (temp) {
                            namelist[active][namesize[active]++]=temp;
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
            if (trap) {
              section="";
              sectnow=true;
              trap=false;
            }
          } while (!feof(handle));
          fclose(handle);
          //do the renaming
          bool doneany=false;
          for (int s=1;s<501;s++) {
            XString newname;
            int k,v;
            if (slist[s]) {
              k=shipdata[s].wShipType;
              if (namesize[k]<=0) {
                if (namesize[k]==0) {
                  fprintf(stderr,"Exhausted names for %s.\n",
                    (const char *)hullspec[shipdata[s].wShipType].sName);
                  namesize[k]=-1;
                }
                k=0;
                if (namesize[k]<=0) {
                  fprintf(stderr,"Exhausted generic names.\n");
                  namesize[k]=-1;
                  k=-1;
                }
              }
              if (k>=0) {
                v=random() % namesize[k];
                newname=namelist[k][v];
                newname=newname.Left(20);
                namesize[k]--;
                namelist[k][v]=namelist[k][namesize[k]];
                namelist[k][namesize[k]]="";
                printf("Renaming %s to %s\n",(const char *)shipdata[s].sName,(const char *)newname);
                shipdata[s].sName=newname;
                write_shipdata(s,&shipdata[s]);
                doneany=true;
              }
            }
          }
          if (doneany) make_checksum(true,false,false);
        }
      }

      delete[] shipdata;
      delete[] truehull;
    }
  }

  delete[] hullspec;

  return 0;
}
